from __future__ import annotations

import os
import subprocess
import sys
from functools import cache

from duty._internal.tools._base import Tool


@cache
def _find_ruff() -> str:
    from ruff.__main__ import find_ruff_bin  # noqa: PLC0415

    try:
        return find_ruff_bin()
    except FileNotFoundError:
        paths = os.environ["PATH"]
        for path in paths.split(os.pathsep):
            ruff = os.path.join(path, "ruff")
            if os.path.exists(ruff):
                return ruff
        py_version = f"{sys.version_info[0]}.{sys.version_info[1]}"
        pypackages_bin = os.path.join("__pypackages__", py_version, "bin")
        ruff = os.path.join(pypackages_bin, "ruff")
        if os.path.exists(ruff):
            return ruff
    return "ruff"


class ruff(Tool):  # noqa: N801
    """Call [Ruff](https://github.com/astral-sh/ruff)."""

    cli_name = "ruff"
    """The name of the executable on PATH."""

    @classmethod
    def check(
        cls,
        *files: str,
        config: str | None = None,
        fix: bool | None = None,
        show_source: bool | None = None,
        show_fixes: bool | None = None,
        diff: bool | None = None,
        watch: bool | None = None,
        fix_only: bool | None = None,
        output_format: str | None = None,
        statistics: bool | None = None,
        add_noqa: bool | None = None,
        show_files: bool | None = None,
        show_settings: bool | None = None,
        select: list[str] | None = None,
        ignore: list[str] | None = None,
        extend_select: list[str] | None = None,
        per_file_ignores: dict[str, list[str]] | None = None,
        fixable: list[str] | None = None,
        unfixable: list[str] | None = None,
        exclude: list[str] | None = None,
        extend_exclude: list[str] | None = None,
        respect_gitignore: bool | None = None,
        force_exclude: bool | None = None,
        no_cache: bool | None = None,
        isolated: bool | None = None,
        cache_dir: str | None = None,
        stdin_filename: str | None = None,
        exit_zero: bool | None = None,
        exit_non_zero_on_fix: bool | None = None,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """Run Ruff on the given files or directories.

        Parameters:
            fix: Attempt to automatically fix lint violations.
            config: Path to the `pyproject.toml` or `ruff.toml` file to use for configuration.
            show_source: Show violations with source code.
            show_fixes: Show an enumeration of all autofixed lint violations.
            diff: Avoid writing any fixed files back; instead, output a diff for each changed file to stdout.
            watch: Run in watch mode by re-running whenever files change.
            fix_only: Fix any fixable lint violations, but don't report on leftover violations. Implies `--fix`.
            output_format: Output serialization format for violations (env: RUFF_FORMAT=) (possible values: text, json, junit, grouped, github, gitlab, pylint).
            statistics: Show counts for every rule with at least one violation.
            add_noqa: Enable automatic additions of `noqa` directives to failing lines.
            show_files: See the files Ruff will be run against with the current settings.
            show_settings: See the settings Ruff will use to lint a given Python file.
            select: Comma-separated list of rule codes to enable (or ALL, to enable all rules).
            ignore: Comma-separated list of rule codes to disable.
            extend_select: Like --select, but adds additional rule codes on top of the selected ones.
            per_file_ignores: List of mappings from file pattern to code to exclude.
            fixable: List of rule codes to treat as eligible for autofix. Only applicable when autofix itself is enabled (e.g., via `--fix`).
            unfixable: List of rule codes to treat as ineligible for autofix. Only applicable when autofix itself is enabled (e.g., via `--fix`).
            exclude: List of paths, used to omit files and/or directories from analysis.
            extend_exclude: Like --exclude, but adds additional files and directories on top of those already excluded.
            respect_gitignore: Respect file exclusions via `.gitignore` and other standard ignore files.
            force_exclude: Enforce exclusions, even for paths passed to Ruff directly on the command-line.
            no_cache: Disable cache reads.
            isolated: Ignore all configuration files.
            cache_dir: Path to the cache directory (env: RUFF_CACHE_DIR=).
            stdin_filename: The name of the file when passing it through stdin.
            exit_zero: Exit with status code "0", even upon detecting lint violations.
            exit_non_zero_on_fix: Exit with a non-zero status code if any files were modified via autofix, even if no lint violations remain.
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["check", *files]

        if fix:
            cli_args.append("--fix")

        if show_source:
            cli_args.append("--show-source")

        if show_fixes:
            cli_args.append("--show-fixes")

        if diff:
            cli_args.append("--diff")

        if watch:
            cli_args.append("--watch")

        if fix_only:
            cli_args.append("--fix-only")

        if output_format:
            cli_args.append("--format")
            cli_args.append(output_format)

        if config:
            cli_args.append("--config")
            cli_args.append(config)

        if statistics:
            cli_args.append("--statistics")

        if add_noqa:
            cli_args.append("--add-noqa")

        if show_files:
            cli_args.append("--show-files")

        if show_settings:
            cli_args.append("--show-settings")

        if select:
            cli_args.append("--select")
            cli_args.append(",".join(select))

        if ignore:
            cli_args.append("--ignore")
            cli_args.append(",".join(ignore))

        if extend_select:
            cli_args.append("--extend-select")
            cli_args.append(",".join(extend_select))

        if per_file_ignores:
            cli_args.append("--per-file-ignores")
            cli_args.append(
                " ".join(f"{path}:{','.join(codes)}" for path, codes in per_file_ignores.items()),
            )

        if fixable:
            cli_args.append("--fixable")
            cli_args.append(",".join(fixable))

        if unfixable:
            cli_args.append("--unfixable")
            cli_args.append(",".join(unfixable))

        if exclude:
            cli_args.append("--exclude")
            cli_args.append(",".join(exclude))

        if extend_exclude:
            cli_args.append("--extend-exclude")
            cli_args.append(",".join(extend_exclude))

        if respect_gitignore:
            cli_args.append("--respect-gitignore")

        if force_exclude:
            cli_args.append("--force-exclude")

        if no_cache:
            cli_args.append("--no-cache")

        if isolated:
            cli_args.append("--isolated")

        if cache_dir:
            cli_args.append("--cache-dir")
            cli_args.append(cache_dir)

        if stdin_filename:
            cli_args.append("--stdin-filename")
            cli_args.append(stdin_filename)

        if exit_zero:
            cli_args.append("--exit-zero")

        if exit_non_zero_on_fix:
            cli_args.append("--exit-non-zero-on-fix")

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    @classmethod
    def format(
        cls,
        *files: str,
        config: str | None = None,
        check: bool | None = None,
        diff: bool | None = None,
        target_version: str | None = None,
        preview: bool | None = None,
        exclude: list[str] | None = None,
        extend_exclude: list[str] | None = None,
        respect_gitignore: bool | None = None,
        force_exclude: bool | None = None,
        no_cache: bool | None = None,
        isolated: bool | None = None,
        cache_dir: str | None = None,
        stdin_filename: str | None = None,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """Run Ruff formatter on the given files or directories.

        Parameters:
            check: Avoid writing any formatted files back; instead, exit with a non-zero status code if any files would have been modified, and zero otherwise
            config: Path to the `pyproject.toml` or `ruff.toml` file to use for configuration
            diff: Avoid writing any fixed files back; instead, output a diff for each changed file to stdout
            target_version: The minimum Python version that should be supported [possible values: py37, py38, py39, py310, py311, py312]
            preview: Enable preview mode; enables unstable formatting
            exclude: List of paths, used to omit files and/or directories from analysis
            extend_exclude: Like --exclude, but adds additional files and directories on top of those already excluded
            respect_gitignore: Respect file exclusions via `.gitignore` and other standard ignore files
            force_exclude: Enforce exclusions, even for paths passed to Ruff directly on the command-line
            no_cache: Disable cache reads
            isolated: Ignore all configuration files
            cache_dir: Path to the cache directory [env: RUFF_CACHE_DIR=]
            stdin_filename: The name of the file when passing it through stdin
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["format", *files]

        if check:
            cli_args.append("--check")

        if diff:
            cli_args.append("--diff")

        if config:
            cli_args.append("--config")
            cli_args.append(config)

        if target_version:
            cli_args.append("--target-version")
            cli_args.append(target_version)

        if preview:
            cli_args.append("--preview")

        if exclude:
            cli_args.append("--exclude")
            cli_args.append(",".join(exclude))

        if extend_exclude:
            cli_args.append("--extend-exclude")
            cli_args.append(",".join(extend_exclude))

        if respect_gitignore:
            cli_args.append("--respect-gitignore")

        if force_exclude:
            cli_args.append("--force-exclude")

        if no_cache:
            cli_args.append("--no-cache")

        if isolated:
            cli_args.append("--isolated")

        if cache_dir:
            cli_args.append("--cache-dir")
            cli_args.append(cache_dir)

        if stdin_filename:
            cli_args.append("--stdin-filename")
            cli_args.append(stdin_filename)

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    @classmethod
    def rule(
        cls,
        rule: str,
        *,
        output_format: str | None = None,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """Explain a rule.

        Parameters:
            rule: A rule code, or `--all`.
            output_format: Output format (default: pretty, possible values: text, json, pretty).
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["rule", rule]

        if output_format:
            cli_args.append("--format")
            cli_args.append(output_format)

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    @classmethod
    def config(
        cls,
        *,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """List or describe the available configuration options.

        Parameters:
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["config"]

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    @classmethod
    def linter(
        cls,
        *,
        output_format: str | None = None,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """List all supported upstream linters.

        Parameters:
            output_format: Output format (default: pretty) (possible values: text, json, pretty).
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["linter"]

        if output_format:
            cli_args.append("--format")
            cli_args.append(output_format)

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    @classmethod
    def clean(
        cls,
        *,
        verbose: bool = False,
        quiet: bool = False,
        silent: bool = False,
    ) -> ruff:
        """Clear any caches in the current directory and any subdirectories.

        Parameters:
            verbose: Enable verbose logging.
            quiet: Print lint violations, but nothing else.
            silent: Disable all logging (but still exit with status code "1" upon detecting lint violations).
        """
        cli_args = ["clean"]

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if silent:
            cli_args.append("--silent")

        return cls(cli_args)

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        process = subprocess.run(  # noqa: S603
            [_find_ruff(), *self.cli_args],
            capture_output=True,
            text=True,
            check=False,
        )
        print(process.stdout)  # noqa: T201
        return process.returncode



"""Tests for the `validation` module."""

from __future__ import annotations

from inspect import Parameter
from typing import Any, Callable

import pytest

from duty._internal.validation import _get_params_caster, cast_arg, to_bool
from tests.fixtures import validation as valfix


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("y", True),
        ("Y", True),
        ("yes", True),
        ("YES", True),
        ("on", True),
        ("ON", True),
        ("true", True),
        ("TRUE", True),
        ("anything else", True),
        ("-1", True),
        ("1", True),
        ("", False),
        ("n", False),
        ("N", False),
        ("no", False),
        ("NO", False),
        ("false", False),
        ("FALSE", False),
        ("off", False),
        ("OFF", False),
    ],
)
def test_bool_casting(value: str, expected: bool) -> None:
    """Check that we correctly cast string values to booleans.

    Parameters:
        value: The value to cast.
        expected: The expected result.
    """
    assert to_bool(value) == expected


class CustomType1:  # noqa: PLW1641
    """Dummy type to test type-casting."""

    def __init__(self, value: str):  # noqa: D107
        self.value = value

    def __eq__(self, other: object):
        return self.value == other.value  # type: ignore[attr-defined]


class CustomType2:
    """Dummy type to test type-casting."""

    def __init__(self, value, extra):  # noqa: ANN001,D107
        ...  # pragma: no cover


@pytest.mark.parametrize(
    ("arg", "annotation", "expected"),
    [
        ("hello", Parameter.empty, "hello"),
        ("off", bool, False),
        ("on", bool, True),
        ("1", int, 1),
        ("1", float, 1.0),
        ("fie", str, "fie"),
        ("fih", CustomType1, CustomType1("fih")),
        ("foh", CustomType2, "foh"),
    ],
)
def test_cast_arg(arg: str, annotation: Any, expected: Any) -> None:
    """Check that arguments are properly casted given an annotation.

    Parameters:
        arg: The argument value to cast.
        annotation: The annotation to use.
        expected: The expected result.
    """
    assert cast_arg(arg, annotation) == expected


_parametrization = [
    (valfix.no_params, (), {}, (), {}),
    (valfix.pos_or_kw_param, ("1",), {}, (1,), {}),
    (valfix.pos_or_kw_param, (), {"a": "1"}, (), {"a": 1}),
    (valfix.pos_or_kw_params, ("1", "2"), {}, (1, 2), {}),
    (valfix.pos_or_kw_params, ("1",), {"b": "2"}, (1,), {"b": 2}),
    (valfix.pos_or_kw_params, (), {"a": "1", "b": "2"}, (), {"a": 1, "b": 2}),
    (valfix.varpos_param, (), {}, (), {}),
    (valfix.varpos_param, ("1", "2"), {}, (1, 2), {}),
    (valfix.pos_and_varpos_param, ("1",), {}, (1,), {}),
    (valfix.pos_and_varpos_param, ("1", "2"), {}, (1, 2), {}),
    (valfix.pos_and_varpos_param, ("1", "2", "3"), {}, (1, 2, 3), {}),
    (valfix.kwonly_param, (), {"b": "1"}, (), {"b": 1}),
    (valfix.kwonly_param, ("2",), {"b": "1"}, (2,), {"b": 1}),
    (valfix.kwonly_param, ("2", "3"), {"b": "1"}, (2, 3), {"b": 1}),
    (valfix.varkw_param, ("1",), {}, (1,), {}),
    (valfix.varkw_param, ("1",), {"b": "2"}, (1,), {"b": 2}),
    (valfix.varkw_param, ("1",), {"b": "2", "c": "3"}, (1,), {"b": 2, "c": 3}),
    (valfix.varkw_no_annotation, (), {"a": "1"}, (), {"a": "1"}),
    (valfix.posonly_marker, ("1", "2"), {}, (1, 2), {}),
    (valfix.posonly_marker, ("1",), {"b": "2"}, (1,), {"b": 2}),
    (valfix.kwonly_marker, ("1",), {"b": "2"}, (1,), {"b": 2}),
    (valfix.kwonly_marker, (), {"a": "1", "b": "2"}, (), {"a": 1, "b": 2}),
    (valfix.only_markers, ("1",), {"b": "2", "c": "3"}, (1,), {"b": 2, "c": 3}),
    (valfix.only_markers, ("1", "2"), {"c": "3"}, (1, 2), {"c": 3}),
    (valfix.full, ("1", "2", "3", "4"), {"d": "5", "e": "6", "f": "7"}, (1, 2, 3, 4), {"d": 5, "e": 6, "f": 7}),
]


@pytest.mark.parametrize(
    ("func", "args", "kwargs", "expected_args", "expected_kwargs"),
    _parametrization,
)
def test_params_caster(func: Callable, args: tuple, kwargs: dict, expected_args: tuple, expected_kwargs: dict) -> None:
    """Test the whole parameters casting helper class.

    Parameters:
        func: The function to work with.
        args: The positional arguments to cast.
        kwargs: The keyword arguments to cast.
        expected_args: The expected positional arguments result.
        expected_kwargs: The expected keyword arguments result.
    """
    caster = _get_params_caster(func, *args, **kwargs)
    new_args, new_kwargs = caster.cast(*args, **kwargs)
    assert new_args == expected_args
    assert new_kwargs == expected_kwargs


def test_casting_based_on_default_value_type() -> None:
    """Test that we cast according to the default value type when there is no annotation."""

    def func(ctx, a=0):  # noqa: ANN202, ANN001
        ...

    caster = _get_params_caster(func, a="1")
    _, kwargs = caster.cast(a="1")
    assert kwargs == {"a": 1}


def test_validating_modern_annotations() -> None:
    """Test modern type annotations in function signatures."""

    def func(ctx, a: int | None = None):  # noqa: ANN202, ANN001
        ...

    caster = _get_params_caster(func, a=1)
    _, kwargs = caster.cast(a="1")
    assert kwargs == {"a": 1}
    caster = _get_params_caster(func, a=None)
    _, kwargs = caster.cast(a=None)
    assert kwargs == {"a": None}
    caster = _get_params_caster(func)
    _, kwargs = caster.cast()
    assert kwargs == {}



# YORE: Bump 2: Remove file.

from __future__ import annotations

from typing import Literal

from failprint import lazy

_BADGE_STYLE = Literal["flat", "flat-square", "flat-square-modified", "for-the-badge", "plastic", "social"]


@lazy(name="interrogate")
def run(
    *src: str,
    verbose: int | None = None,
    quiet: bool | None = None,
    fail_under: float | None = None,
    exclude: str | None = None,
    ignore_init_method: bool | None = None,
    ignore_init_module: bool | None = None,
    ignore_magic: bool | None = None,
    ignore_module: bool | None = None,
    ignore_nested_functions: bool | None = None,
    ignore_nested_classes: bool | None = None,
    ignore_private: bool | None = None,
    ignore_property_decorators: bool | None = None,
    ignore_setters: bool | None = None,
    ignore_semiprivate: bool | None = None,
    ignore_regex: str | None = None,
    whitelist_regex: str | None = None,
    output: str | None = None,
    config: str | None = None,
    color: bool | None = None,
    omit_covered_files: bool | None = None,
    generate_badge: str | None = None,
    badge_format: Literal["png", "svg"] | None = None,
    badge_style: _BADGE_STYLE | None = None,
) -> None:
    """Run `interrogate`.

    Args:
        src: Format the directories and file paths.
        verbose: Level of verbosity.
        quiet: Do not print output.
        fail_under: Fail when coverage % is less than a given amount.
        exclude: Exclude PATHs of files and/or directories.
        ignore_init_method: Ignore `__init__` method of classes.
        ignore_init_module: Ignore `__init__.py` modules.
        ignore_magic: Ignore all magic methods of classes.
        ignore_module: Ignore module-level docstrings.
        ignore_nested_functions: Ignore nested functions and methods.
        ignore_nested_classes: Ignore nested classes.
        ignore_private: Ignore private classes, methods, and functions starting with two underscores.
        ignore_property_decorators: Ignore methods with property setter/getter decorators.
        ignore_setters: Ignore methods with property setter decorators.
        ignore_semiprivate: Ignore semiprivate classes, methods, and functions starting with a single underscore.
        ignore_regex: Regex identifying class, method, and function names to ignore.
        whitelist_regex: Regex identifying class, method, and function names to include.
        output: Write output to a given FILE.
        config: Read configuration from pyproject.toml or setup.cfg.
        color: Toggle color output on/off when printing to stdout.
        omit_covered_files: Omit reporting files that have 100% documentation coverage.
        generate_badge: Generate a shields.io status badge (an SVG image) in at a given file or directory.
        badge_format: File format for the generated badge.
        badge_style: Desired style of shields.io badge.
    """
    from interrogate.cli import main as interrogate  # noqa: PLC0415

    cli_args: list[str] = list(src)

    if verbose:
        cli_args.append("--verbose")
        cli_args.append(str(verbose))

    if quiet:
        cli_args.append("--quiet")

    if fail_under:
        cli_args.append("--fail-under")
        cli_args.append(str(fail_under))

    if exclude:
        cli_args.append("--exclude")
        cli_args.append(exclude)

    if ignore_init_method:
        cli_args.append("--ignore-init-method")

    if ignore_init_module:
        cli_args.append("--ignore-init-module")

    if ignore_magic:
        cli_args.append("--ignore-magic")

    if ignore_module:
        cli_args.append("--ignore-module")

    if ignore_nested_functions:
        cli_args.append("--ignore-nested-functions")

    if ignore_nested_classes:
        cli_args.append("--ignore-nested-classes")

    if ignore_private:
        cli_args.append("--ignore-private")

    if ignore_property_decorators:
        cli_args.append("--ignore-property-decorators")

    if ignore_setters:
        cli_args.append("--ignore-setters")

    if ignore_semiprivate:
        cli_args.append("--ignore-semiprivate")

    if ignore_regex:
        cli_args.append("--ignore-regex")
        cli_args.append(ignore_regex)

    if whitelist_regex:
        cli_args.append("--whitelist-regex")
        cli_args.append(whitelist_regex)

    if output:
        cli_args.append("--output")
        cli_args.append(output)

    if omit_covered_files:
        cli_args.append("--omit-covered-files")

    if generate_badge:
        cli_args.append("--generate-badge")
        cli_args.append(generate_badge)

    if badge_format:
        cli_args.append("--badge-format")
        cli_args.append(badge_format)

    if badge_style:
        cli_args.append("--badge-style")
        cli_args.append(badge_style)

    if config:
        cli_args.append("--config")
        cli_args.append(config)

    if color is True:
        cli_args.append("--color")
    elif color is False:
        cli_args.append("--no-color")

    interrogate(cli_args)



from __future__ import annotations

from duty._internal.tools._base import LazyStderr, LazyStdout, Tool


class autoflake(Tool):  # noqa: N801
    """Call [autoflake](https://github.com/PyCQA/autoflake)."""

    cli_name = "autoflake"
    """The name of the executable on PATH."""

    def __init__(
        self,
        *files: str,
        config: str | None = None,
        check: bool | None = None,
        check_diff: bool | None = None,
        imports: list[str] | None = None,
        remove_all_unused_imports: bool | None = None,
        recursive: bool | None = None,
        jobs: int | None = None,
        exclude: list[str] | None = None,
        expand_star_imports: bool | None = None,
        ignore_init_module_imports: bool | None = None,
        remove_duplicate_keys: bool | None = None,
        remove_unused_variables: bool | None = None,
        remove_rhs_for_unused_variables: bool | None = None,
        ignore_pass_statements: bool | None = None,
        ignore_pass_after_docstring: bool | None = None,
        quiet: bool | None = None,
        verbose: bool | None = None,
        stdin_display_name: str | None = None,
        in_place: bool | None = None,
        stdout: bool | None = None,
    ) -> None:
        r"""Run `autoflake`.

        Parameters:
            *files: Files to format.
            config: Explicitly set the config file instead of auto determining based on file location.
            check: Return error code if changes are needed.
            check_diff: Return error code if changes are needed, also display file diffs.
            imports: By default, only unused standard library imports are removed; specify a comma-separated list of additional modules/packages.
            remove_all_unused_imports: Remove all unused imports (not just those from the standard library).
            recursive: Drill down directories recursively.
            jobs: Number of parallel jobs; match CPU count if value is 0 (default: 0).
            exclude: Exclude file/directory names that match these comma-separated globs.
            expand_star_imports: Expand wildcard star imports with undefined names; this only triggers if there is only one star import in the file; this is skipped if there are any uses of `__all__` or `del` in the file.
            ignore_init_module_imports: Exclude `__init__.py` when removing unused imports.
            remove_duplicate_keys: Remove all duplicate keys in objects.
            remove_unused_variables: Remove unused variables.
            remove_rhs_for_unused_variables: Remove RHS of statements when removing unused variables (unsafe).
            ignore_pass_statements: Ignore all pass statements.
            ignore_pass_after_docstring: Ignore pass statements after a newline ending on `\"\"\"`.
            quiet: Suppress output if there are no issues.
            verbose: Print more verbose logs (you can repeat `-v` to make it more verbose).
            stdin_display_name: The name used when processing input from stdin.
            in_place: Make changes to files instead of printing diffs.
            stdout: Print changed text to stdout. defaults to true when formatting stdin, or to false otherwise.
        """
        cli_args = list(files)

        if check:
            cli_args.append("--check")

        if check_diff:
            cli_args.append("--check-diff")

        if imports:
            cli_args.append("--imports")
            cli_args.append(",".join(imports))

        if remove_all_unused_imports:
            cli_args.append("--remove-all-unused-imports")

        if recursive:
            cli_args.append("--recursive")

        if jobs:
            cli_args.append("--jobs")
            cli_args.append(str(jobs))

        if exclude:
            cli_args.append("--exclude")
            cli_args.append(",".join(exclude))

        if expand_star_imports:
            cli_args.append("--expand-star-imports")

        if ignore_init_module_imports:
            cli_args.append("--ignore-init-module-imports")

        if remove_duplicate_keys:
            cli_args.append("--remove-duplicate-keys")

        if remove_unused_variables:
            cli_args.append("--remove-unused-variables")

        if remove_rhs_for_unused_variables:
            cli_args.append("remove-rhs-for-unused-variables")

        if ignore_pass_statements:
            cli_args.append("--ignore-pass-statements")

        if ignore_pass_after_docstring:
            cli_args.append("--ignore-pass-after-docstring")

        if quiet:
            cli_args.append("--quiet")

        if verbose:
            cli_args.append("--verbose")

        if stdin_display_name:
            cli_args.append("--stdin-display-name")
            cli_args.append(stdin_display_name)

        if config:
            cli_args.append("--config")
            cli_args.append(config)

        if in_place:
            cli_args.append("--in-place")

        if stdout:
            cli_args.append("--stdout")

        super().__init__(cli_args)

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        from autoflake import _main as run_autoflake  # noqa: PLC0415

        return run_autoflake(
            self.cli_args,
            standard_out=LazyStdout(),
            standard_error=LazyStderr(),
        )



from __future__ import annotations

from typing import Literal

from duty._internal.tools._base import Tool


class build(Tool):  # noqa: N801
    """Call [build](https://github.com/pypa/build)."""

    cli_name = "pyproject-build"
    """The name of the executable on PATH."""

    def __init__(
        self,
        srcdir: str | None = None,
        *,
        version: bool = False,
        verbose: bool = False,
        sdist: bool = False,
        wheel: bool = False,
        outdir: str | None = None,
        skip_dependency_check: bool = False,
        no_isolation: bool = False,
        installer: Literal["pip", "uv"] | None = None,
        config_setting: list[str] | None = None,
    ) -> None:
        """Run `build`.

        Parameters:
            srcdir: Source directory (defaults to current directory).
            version: Show program's version number and exit.
            verbose: Increase verbosity
            sdist: Build a source distribution (disables the default behavior).
            wheel: Build a wheel (disables the default behavior).
            outdir: Output directory (defaults to `{srcdir}/dist`).
            skip_dependency_check: Do not check that build dependencies are installed.
            no_isolation: Disable building the project in an isolated virtual environment.
                Build dependencies must be installed separately when this option is used.
            installer: Python package installer to use (defaults to pip).
            config_setting: Settings to pass to the backend. Multiple settings can be provided.
        """
        cli_args = []

        if srcdir:
            cli_args.append(srcdir)

        if version:
            cli_args.append("--version")

        if verbose:
            cli_args.append("--verbose")

        if sdist:
            cli_args.append("--sdist")

        if wheel:
            cli_args.append("--wheel")

        if outdir:
            cli_args.append("--outdir")
            cli_args.append(outdir)

        if skip_dependency_check:
            cli_args.append("--skip-dependency-check")

        if no_isolation:
            cli_args.append("--no-isolation")

        if installer:
            cli_args.append("--installer")
            cli_args.append(installer)

        if config_setting:
            for setting in config_setting:
                cli_args.append(f"--config-setting={setting}")

        super().__init__(cli_args)

    def __call__(self) -> None:
        """Run the command."""
        from build.__main__ import main as run_build  # noqa: PLC0415

        run_build(self.cli_args)



from __future__ import annotations

from typing import Literal

from duty._internal.tools._base import Tool


class git_changelog(Tool):  # noqa: N801
    """Call [git-changelog](https://github.com/pawamoy/git-changelog)."""

    cli_name = "git-changelog"
    """The name of the executable on PATH."""

    def __init__(
        self,
        repository: str | None = None,
        *,
        config_file: str | None = None,
        bump: str | None = None,
        versioning: Literal["semver", "pep440"] | None = None,
        in_place: bool = False,
        version_regex: str | None = None,
        marker_line: str | None = None,
        output: str | None = None,
        provider: Literal["github", "gitlab", "bitbucket"] | None = None,
        parse_refs: bool = False,
        release_notes: bool = False,
        input: str | None = None,  # noqa: A002
        convention: Literal["basic", "angular", "conventional"] | None = None,
        sections: list[str] | None = None,
        template: str | None = None,
        git_trailers: bool = False,
        omit_empty_versions: bool = False,
        no_zerover: bool = False,
        filter_commits: str | None = None,
        jinja_context: list[str] | None = None,
        version: bool = False,
        debug_info: bool = False,
    ) -> None:
        r"""Run `git-changelog`.

        Parameters:
            repository: The repository path, relative or absolute. Default: current working directory.
            config_file: Configuration file(s).
            bump: Specify the bump from latest version for the set of unreleased commits.
                Can be one of `auto`, `major`, `minor`, `patch` or a valid SemVer version (eg. 1.2.3).
                For both SemVer and PEP 440 versioning schemes (see -n), `auto` will bump the major number
                if a commit contains breaking changes (or the minor number for 0.x versions, see -Z),
                else the minor number if there are new features, else the patch number. Default: unset (false).
            versioning: Versioning scheme to use when bumping and comparing versions.
                The selected scheme will impact the values accepted by the `--bump` option.
                Supported: `pep440`, `semver`.

                PEP 440 provides the following bump strategies: `auto`, `epoch`, `release`, `major`, `minor`, `micro`, `patch`,
                `pre`, `alpha`, `beta`, `candidate`, `post`, `dev`.
                Values `auto`, `major`, `minor`, `micro` can be suffixed with one of `+alpha`, `+beta`, `+candidate`, and/or `+dev`.
                Values `alpha`, `beta` and `candidate` can be suffixed with `+dev`.
                Examples: `auto+alpha`, `major+beta+dev`, `micro+dev`, `candidate+dev`, etc..

                SemVer provides the following bump strategies: `auto`, `major`, `minor`, `patch`, `release`.
                See the docs for more information. Default: unset (`semver`).
            in_place: Insert new entries (versions missing from changelog) in-place.
                An output file must be specified. With custom templates, you can pass two additional
                arguments: `--version-regex` and `--marker-line`.
                When writing in-place, an `in_place` variable will be injected in the Jinja context,
                allowing to adapt the generated contents (for example to skip changelog headers or footers).
                Default: unset (false).
            version_regex: A regular expression to match versions in the existing changelog
                (used to find the latest release) when writing in-place.
                The regular expression must be a Python regex with a `version` named group.
                Default: `^## \[(?P<version>v?[^\]]+)`.
            marker_line: A marker line at which to insert new entries (versions missing from changelog).
                If two marker lines are present in the changelog, the contents between those two lines
                will be overwritten (useful to update an 'Unreleased' entry for example). Default: `<!-- insertion marker -->`.
            output: Output to given file. Default: standard output.
            provider: Explicitly specify the repository provider. Default: unset.
            parse_refs: Parse provider-specific references in commit messages (GitHub/GitLab/Bitbucket issues, PRs, etc.).
                Default: unset (false).
            release_notes: Output release notes to stdout based on the last entry in the changelog. Default: unset (false).
            input: Read from given file when creating release notes. Default: `CHANGELOG.md`.
            convention: The commit convention to match against. Default: `basic`.
            sections: A comma-separated list of sections to render.
                See the available sections for each supported convention in the description. Default: unset (None).
            template: The Jinja2 template to use.
                Prefix it with `path:` to specify the path to a Jinja templated file. Default: `keepachangelog`.
            git_trailers: Parse Git trailers in the commit message.
                See https://git-scm.com/docs/git-interpret-trailers. Default: unset (false).
            omit_empty_versions: Omit empty versions from the output. Default: unset (false).
            no_zerover: By default, breaking changes on a 0.x don't bump the major version, maintaining it at 0.
                With this option, a breaking change will bump a 0.x version to 1.0.
            filter_commits: The Git revision-range filter to use (e.g. `v1.2.0..`). Default: no filter.
            jinja_context: Pass additional key/value pairs to the template.
                Option can be used multiple times.
                The key/value pairs are accessible as 'jinja_context' in the template.
            version: Show the current version of the program and exit.
            debug_info: Print debug information.
        """
        cli_args = []

        if repository:
            cli_args.append(repository)

        if config_file:
            cli_args.append("--config-file")
            cli_args.append(config_file)

        if bump:
            cli_args.append("--bump")
            cli_args.append(bump)

        if versioning:
            cli_args.append("--versioning")
            cli_args.append(versioning)

        if in_place:
            cli_args.append("--in-place")

        if version_regex:
            cli_args.append("--version-regex")
            cli_args.append(version_regex)

        if marker_line:
            cli_args.append("--marker-line")
            cli_args.append(marker_line)

        if output:
            cli_args.append("--output")
            cli_args.append(output)

        if provider:
            cli_args.append("--provider")
            cli_args.append(provider)

        if parse_refs:
            cli_args.append("--parse-refs")

        if release_notes:
            cli_args.append("--release-notes")

        if input:
            cli_args.append("--input")
            cli_args.append(input)

        if convention:
            cli_args.append("--convention")
            cli_args.append(convention)

        if sections:
            cli_args.append("--sections")
            cli_args.append(",".join(sections))

        if template:
            cli_args.append("--template")
            cli_args.append(template)

        if git_trailers:
            cli_args.append("--git-trailers")

        if omit_empty_versions:
            cli_args.append("--omit-empty-versions")

        if no_zerover:
            cli_args.append("--no-zerover")

        if filter_commits:
            cli_args.append("--filter-commits")
            cli_args.append(filter_commits)

        if jinja_context:
            for key_value in jinja_context:
                cli_args.append("--jinja-context")
                cli_args.append(key_value)

        if version:
            cli_args.append("--version")

        if debug_info:
            cli_args.append("--debug-info")

        super().__init__(cli_args)

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        from git_changelog import main as run_git_changelog  # noqa: PLC0415

        return run_git_changelog(self.cli_args)



# YORE: Bump 2: Remove file.

from __future__ import annotations

from failprint import lazy


def run(*args: str, version: bool = False, no_color: bool = False) -> None:
    """Run `twine`.

    Parameters:
        *args: CLI arguments.
        version: Show program's version number and exit.
        no_color: Disable colored output.
    """
    from twine.cli import dispatch as twine  # noqa: PLC0415

    cli_args = []

    # --version and --no-color must appear first.
    if version:
        cli_args.append("--version")

    if no_color:
        cli_args.append("--no-color")

    cli_args.extend(args)

    twine(cli_args)


@lazy(name="twine.check")
def check(
    *dists: str,
    strict: bool = False,
    version: bool = False,
    no_color: bool = False,
) -> None:
    """Checks whether your distribution's long description will render correctly on PyPI.

    Parameters:
        dists: The distribution files to check, usually `dist/*`.
        strict: Fail on warnings.
        version: Show program's version number and exit.
        no_color: Disable colored output.
    """
    cli_args = list(dists)

    if strict is True:
        cli_args.append("--strict")

    run("check", *cli_args, version=version, no_color=no_color)


@lazy(name="twine.register")
def register(
    package: str,
    *,
    repository: str = "pypi",
    repository_url: str | None = None,
    attestations: bool = False,
    sign: bool = False,
    sign_with: str | None = None,
    identity: str | None = None,
    username: str | None = None,
    password: str | None = None,
    non_interactive: bool = False,
    comment: str | None = None,
    config_file: str | None = None,
    skip_existing: bool = False,
    cert: str | None = None,
    client_cert: str | None = None,
    verbose: bool = False,
    disable_progress_bar: bool = False,
    version: bool = False,
    no_color: bool = False,
) -> None:
    """Pre-register a name with a repository before uploading a distribution.

    Pre-registration is not supported on PyPI, so the register command
    is only necessary if you are using a different repository that requires it.

    Parameters:
        package: File from which we read the package metadata.
        repository: The repository (package index) to upload the package to.
            Should be a section in the config file (default: `pypi`).
            Can also be set via `TWINE_REPOSITORY` environment variable.
        repository_url: The repository (package index) URL to upload the package to. This overrides `--repository`.
            Can also be set via `TWINE_REPOSITORY_URL` environment variable.
        attestations: Upload each file's associated attestations.
        sign: Sign files to upload using GPG.
        sign_with: GPG program used to sign uploads (default: `gpg`).
        identity: GPG identity used to sign files.
        username: The username to authenticate to the repository (package index) as.
            Can also be set via `TWINE_USERNAME` environment variable.
        password: The password to authenticate to the repository (package index) with.
            Can also be set via `TWINE_PASSWORD` environment variable.
        non_interactive: Do not interactively prompt for username/password if the required credentials are missing.
            Can also be set via `TWINE_NON_INTERACTIVE` environment variable.
        comment: The comment to include with the distribution file.
        config_file: The `.pypirc` config file to use.
        skip_existing: Continue uploading files if one already exists.
            Only valid when uploading to PyPI. Other implementations may not support this.
        cert: Path to alternate CA bundle (can also be set via `TWINE_CERT` environment variable).
        client_cert: Path to SSL client certificate, a single file containing the private key and the certificate in PEM format.
        verbose: Show verbose output.
        disable_progress_bar: Disable the progress bar.
        version: Show program's version number and exit.
        no_color: Disable colored output.
    """
    cli_args = [package]

    if repository:
        cli_args.append("--repository")
        cli_args.append(repository)

    if repository_url:
        cli_args.append("--repository-url")
        cli_args.append(repository_url)

    if attestations:
        cli_args.append("--attestations")

    if sign:
        cli_args.append("--sign")

    if sign_with:
        cli_args.append("--sign-with")
        cli_args.append(sign_with)

    if identity:
        cli_args.append("--identity")
        cli_args.append(identity)

    if username:
        cli_args.append("--username")
        cli_args.append(username)

    if password:
        cli_args.append("--password")
        cli_args.append(password)

    if non_interactive:
        cli_args.append("--non-interactive")

    if comment:
        cli_args.append("--repository")

    if config_file:
        cli_args.append("--config-file")
        cli_args.append(config_file)

    if skip_existing:
        cli_args.append("--skip-existing")

    if cert:
        cli_args.append("--cert")
        cli_args.append(cert)

    if client_cert:
        cli_args.append("--client-cert")
        cli_args.append(client_cert)

    if verbose:
        cli_args.append("--verbose")

    if disable_progress_bar:
        cli_args.append("--disable-progress-bar")

    run("check", *cli_args, version=version, no_color=no_color)


@lazy(name="twine.upload")
def upload(
    *dists: str,
    repository: str = "pypi",
    repository_url: str | None = None,
    attestations: bool = False,
    sign: bool = False,
    sign_with: str | None = None,
    identity: str | None = None,
    username: str | None = None,
    password: str | None = None,
    non_interactive: bool = False,
    comment: str | None = None,
    config_file: str | None = None,
    skip_existing: bool = False,
    cert: str | None = None,
    client_cert: str | None = None,
    verbose: bool = False,
    disable_progress_bar: bool = False,
    version: bool = False,
    no_color: bool = False,
) -> None:
    """Uploads one or more distributions to a repository.

    Parameters:
        dists: The distribution files to check, usually `dist/*`.
        repository: The repository (package index) to upload the package to.
            Should be a section in the config file (default: `pypi`).
            Can also be set via `TWINE_REPOSITORY` environment variable.
        repository_url: The repository (package index) URL to upload the package to. This overrides `--repository`.
            Can also be set via `TWINE_REPOSITORY_URL` environment variable.
        attestations: Upload each file's associated attestations.
        sign: Sign files to upload using GPG.
        sign_with: GPG program used to sign uploads (default: `gpg`).
        identity: GPG identity used to sign files.
        username: The username to authenticate to the repository (package index) as.
            Can also be set via `TWINE_USERNAME` environment variable.
        password: The password to authenticate to the repository (package index) with.
            Can also be set via `TWINE_PASSWORD` environment variable.
        non_interactive: Do not interactively prompt for username/password if the required credentials are missing.
            Can also be set via `TWINE_NON_INTERACTIVE` environment variable.
        comment: The comment to include with the distribution file.
        config_file: The `.pypirc` config file to use.
        skip_existing: Continue uploading files if one already exists.
            Only valid when uploading to PyPI. Other implementations may not support this.
        cert: Path to alternate CA bundle (can also be set via `TWINE_CERT` environment variable).
        client_cert: Path to SSL client certificate, a single file containing the private key and the certificate in PEM format.
        verbose: Show verbose output.
        disable_progress_bar: Disable the progress bar.
        version: Show program's version number and exit.
        no_color: Disable colored output.
    """
    cli_args = list(dists)

    if repository:
        cli_args.append("--repository")
        cli_args.append(repository)

    if repository_url:
        cli_args.append("--repository-url")
        cli_args.append(repository_url)

    if attestations:
        cli_args.append("--attestations")

    if sign:
        cli_args.append("--sign")

    if sign_with:
        cli_args.append("--sign-with")
        cli_args.append(sign_with)

    if identity:
        cli_args.append("--identity")
        cli_args.append(identity)

    if username:
        cli_args.append("--username")
        cli_args.append(username)

    if password:
        cli_args.append("--password")
        cli_args.append(password)

    if non_interactive:
        cli_args.append("--non-interactive")

    if comment:
        cli_args.append("--repository")

    if config_file:
        cli_args.append("--config-file")
        cli_args.append(config_file)

    if skip_existing:
        cli_args.append("--skip-existing")

    if cert:
        cli_args.append("--cert")
        cli_args.append(cert)

    if client_cert:
        cli_args.append("--client-cert")
        cli_args.append(client_cert)

    if verbose:
        cli_args.append("--verbose")

    if disable_progress_bar:
        cli_args.append("--disable-progress-bar")

    run("upload", *cli_args, version=version, no_color=no_color)



# YORE: Bump 2: Remove file.

from __future__ import annotations

import sys

from failprint import lazy


@lazy(name="ssort")
def run(
    *files: str,
    diff: bool | None = None,
    check: bool | None = None,
) -> int:
    """Run `ssort`.

    Parameters:
        *files: Files to format.
        diff: Prints a diff of all changes ssort would make to a file.
        check: Check the file for unsorted statements. Returns 0 if nothing needs to be changed. Otherwise returns 1.
    """
    from ssort._main import main as ssort  # noqa: PLC0415

    cli_args = list(files)

    if diff:
        cli_args.append("--diff")

    if check:
        cli_args.append("--check")

    old_sys_argv = sys.argv
    sys.argv = ["ssort*", *cli_args]
    try:
        return ssort()
    finally:
        sys.argv = old_sys_argv



"""Tests for the CLI."""

from __future__ import annotations

import pytest

from duty import main
from duty._internal import debug


def test_no_duty(capsys: pytest.CaptureFixture) -> None:
    """Run no duties.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main([]) == 1
    captured = capsys.readouterr()
    assert "Available duties" in captured.out


def test_show_help(capsys: pytest.CaptureFixture) -> None:
    """Show help.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-h"]) == 0
    captured = capsys.readouterr()
    assert "duty" in captured.out


def test_show_help_for_given_duties(capsys: pytest.CaptureFixture) -> None:
    """Show help for given duties.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/basic.py", "-h", "hello"]) == 0
    captured = capsys.readouterr()
    assert "hello" in captured.out


def test_show_help_unknown_duty(capsys: pytest.CaptureFixture) -> None:
    """Show help for an unknown duty.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/basic.py", "-h", "not-here"]) == 0
    captured = capsys.readouterr()
    assert "Unknown duty" in captured.out


def test_select_duties() -> None:
    """Run a duty."""
    assert main(["-d", "tests/fixtures/basic.py", "hello"]) == 0


def test_unknown_duty() -> None:
    """Don't run an unknown duty."""
    assert main(["-d", "tests/fixtures/basic.py", "byebye"]) == 1


def test_incorrect_arguments() -> None:
    """Use incorrect arguments."""
    assert main(["-d", "tests/fixtures/basic.py", "hello=1"]) == 1


# we use 300 because it's slightly above the valid maximum 255
@pytest.mark.parametrize("code", range(-100, 300, 7))
def test_duty_failure(code: int) -> None:
    """Check exit code.

    Parameters:
        code: Code to match.
    """
    assert main(["-d", "tests/fixtures/code.py", "exit_with", f"code={code}"]) == code


def test_multiple_duties(capfd: pytest.CaptureFixture) -> None:
    """Run multiple duties.

    Parameters:
        capfd: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/multiple.py", "first_duty", "second_duty"]) == 0
    captured = capfd.readouterr()
    assert "first" in captured.out
    assert "second" in captured.out


def test_duty_arguments(capfd: pytest.CaptureFixture) -> None:
    """Run duty with arguments.

    Parameters:
        capfd: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/arguments.py", "say_hello", "cat=fabric"]) == 0
    captured = capfd.readouterr()
    assert "cat fabric" in captured.out
    assert "dog dog" in captured.out

    assert main(["-d", "tests/fixtures/arguments.py", "say_hello", "dog=paramiko", "cat=invoke"]) == 0
    captured = capfd.readouterr()
    assert "cat invoke" in captured.out
    assert "dog paramiko" in captured.out


def test_list_duties(capsys: pytest.CaptureFixture) -> None:
    """List duties.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/list.py", "-l"]) == 0
    captured = capsys.readouterr()
    assert "Tong..." in captured.out
    assert "DEUM!" in captured.out


def test_global_options() -> None:
    """Test global options."""
    assert main(["-d", "tests/fixtures/code.py", "-z", "exit_with", "1"]) == 0


def test_global_and_local_options() -> None:
    """Test global and local options."""
    assert main(["-d", "tests/fixtures/code.py", "-z", "exit_with", "-Z", "1"]) == 1


def test_options_precedence() -> None:
    """Test options precedence."""
    # @duty(nofail=True) is overridden by ctx.run(nofail=False)
    assert main(["-d", "tests/fixtures/precedence.py", "precedence"]) == 1

    # ctx.run(nofail=False) is overridden by local option -z
    assert main(["-d", "tests/fixtures/precedence.py", "precedence", "-z"]) == 0

    # ctx.run(nofail=False) is overridden by global option -z
    assert main(["-d", "tests/fixtures/precedence.py", "-z", "precedence"]) == 0

    # global option -z is overridden by local option -z
    assert main(["-d", "tests/fixtures/precedence.py", "-z", "precedence", "-Z"]) == 1


# test options precedence (CLI option, env var, ctx.run, @duty
# test positional arguments
# test extra keyword arguments
# test complete (global options + local options + multi duties + positional args + keyword args + extra keyword args)


@pytest.mark.parametrize(
    ("param", "expected"),
    [
        ("", 1),
        ("n", 1),
        ("N", 1),
        ("no", 1),
        ("NO", 1),
        ("false", 1),
        ("FALSE", 1),
        ("off", 1),
        ("OFF", 1),
        ("zero=", 1),
        ("zero=0", 1),
        ("zero=n", 1),
        ("zero=N", 1),
        ("zero=no", 1),
        ("zero=NO", 1),
        ("zero=false", 1),
        ("zero=FALSE", 1),
        ("zero=off", 1),
        ("zero=OFF", 1),
        ("y", 0),
        ("Y", 0),
        ("yes", 0),
        ("YES", 0),
        ("on", 0),
        ("ON", 0),
        ("true", 0),
        ("TRUE", 0),
        ("anything else", 0),
        ("-1", 0),
        ("1", 0),
        ("zero=y", 0),
        ("zero=Y", 0),
        ("zero=yes", 0),
        ("zero=YES", 0),
        ("zero=on", 0),
        ("zero=ON", 0),
        ("zero=true", 0),
        ("zero=TRUE", 0),
        ("zero=anything else", 0),
        ("zero=-1", 0),
        ("zero=1", 0),
    ],
)
def test_cast_bool_parameter(param: str, expected: int) -> None:
    """Test parameters casting as boolean.

    Parameters:
        param: Pytest parametrization fixture.
        expected: Pytest parametrization fixture.
    """
    assert main(["-d", "tests/fixtures/booleans.py", "boolean", param]) == expected


def test_invalid_params(capsys: pytest.CaptureFixture) -> None:
    """Check that invalid parameters are early and correctly detected.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-d", "tests/fixtures/booleans.py", "boolean", "zore=off"]) == 1
    captured = capsys.readouterr()
    assert "unexpected keyword argument 'zore'" in captured.err

    assert main(["-d", "tests/fixtures/code.py", "exit_with"]) == 1
    captured = capsys.readouterr()
    assert "missing 1 required positional argument: 'code'" in captured.err


def test_show_version(capsys: pytest.CaptureFixture) -> None:
    """Show version.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["-V"])
    captured = capsys.readouterr()
    assert debug._get_version() in captured.out


def test_show_debug_info(capsys: pytest.CaptureFixture) -> None:
    """Show debug information.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["--debug-info"])
    captured = capsys.readouterr().out.lower()
    assert "python" in captured
    assert "system" in captured
    assert "environment" in captured
    assert "packages" in captured



# YORE: Bump 2: Remove file.

from __future__ import annotations

from typing import Literal

from failprint import lazy


@lazy(name="build")
def run(
    srcdir: str | None = None,
    *,
    version: bool = False,
    verbose: bool = False,
    sdist: bool = False,
    wheel: bool = False,
    outdir: str | None = None,
    skip_dependency_check: bool = False,
    no_isolation: bool = False,
    installer: Literal["pip", "uv"] | None = None,
    config_setting: list[str] | None = None,
) -> None:
    """Run `build`.

    Parameters:
        srcdir: Source directory (defaults to current directory).
        version: Show program's version number and exit.
        verbose: Increase verbosity
        sdist: Build a source distribution (disables the default behavior).
        wheel: Build a wheel (disables the default behavior).
        outdir: Output directory (defaults to `{srcdir}/dist`).
        skip_dependency_check: Do not check that build dependencies are installed.
        no_isolation: Disable building the project in an isolated virtual environment.
            Build dependencies must be installed separately when this option is used.
        installer: Python package installer to use (defaults to pip).
        config_setting: Settings to pass to the backend. Multiple settings can be provided.
    """
    from build.__main__ import main as build  # noqa: PLC0415

    cli_args = []

    if srcdir:
        cli_args.append(srcdir)

    if version:
        cli_args.append("--version")

    if verbose:
        cli_args.append("--verbose")

    if sdist:
        cli_args.append("--sdist")

    if wheel:
        cli_args.append("--wheel")

    if outdir:
        cli_args.append("--outdir")
        cli_args.append(outdir)

    if skip_dependency_check:
        cli_args.append("--skip-dependency-check")

    if no_isolation:
        cli_args.append("--no-isolation")

    if installer:
        cli_args.append("--installer")
        cli_args.append(installer)

    if config_setting:
        for setting in config_setting:
            cli_args.append(f"--config-setting={setting}")

    build(cli_args)



from __future__ import annotations

import re
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING

from duty._internal.tools._base import Tool

if TYPE_CHECKING:
    from collections.abc import Sequence


class blacken_docs(Tool):  # noqa: N801
    """Call [blacken-docs](https://github.com/adamchainz/blacken-docs)."""

    cli_name = "blacken-docs"
    """The name of the executable on PATH."""

    def __init__(
        self,
        *paths: str | Path,
        exts: Sequence[str] | None = None,
        exclude: Sequence[str | Pattern] | None = None,
        skip_errors: bool = False,
        rst_literal_blocks: bool = False,
        line_length: int | None = None,
        string_normalization: bool = True,
        is_pyi: bool = False,
        is_ipynb: bool = False,
        skip_source_first_line: bool = False,
        magic_trailing_comma: bool = True,
        python_cell_magics: set[str] | None = None,
        preview: bool = False,
        check_only: bool = False,
    ) -> None:
        """Run `blacken-docs`.

        Parameters:
            *paths: Directories and files to format.
            exts: List of extensions to select files with.
            exclude: List of regular expressions to exclude files.
            skip_errors: Don't exit non-zero for errors from Black (normally syntax errors).
            rst_literal_blocks: Also format literal blocks in reStructuredText files (more below).
            line_length: How many characters per line to allow.
            string_normalization: Normalize string quotes or prefixes.
            is_pyi: Format all input files like typing stubs regardless of file extension.
            is_ipynb: Format all input files like Jupyter Notebooks regardless of file extension.
            skip_source_first_line: Skip the first line of the source code.
            magic_trailing_comma: Use trailing commas as a reason to split lines.
            python_cell_magics: When processing Jupyter Notebooks, add the given magic to the list
                of known python-magics (capture, prun, pypy, python, python3, time, timeit).
                Useful for formatting cells with custom python magics.
            preview: Enable potentially disruptive style changes that may be added
                to Black's main functionality in the next major release.
            check_only: Don't modify files but indicate when changes are necessary
                with a message and non-zero return code.

        Returns:
            Success/failure.
        """
        super().__init__(py_args=dict(locals()))

    @property
    def cli_command(self) -> str:
        """The equivalent CLI command."""
        raise ValueError("This command cannot be translated to a CLI command.")

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        import black  # noqa: PLC0415
        from blacken_docs import format_file  # noqa: PLC0415

        # Restore locals.
        exts = self.py_args["exts"]
        exclude = self.py_args["exclude"]
        paths = self.py_args["paths"]
        line_length = self.py_args["line_length"]
        string_normalization = self.py_args["string_normalization"]
        is_pyi = self.py_args["is_pyi"]
        is_ipynb = self.py_args["is_ipynb"]
        skip_source_first_line = self.py_args["skip_source_first_line"]
        magic_trailing_comma = self.py_args["magic_trailing_comma"]
        python_cell_magics = self.py_args["python_cell_magics"]
        preview = self.py_args["preview"]
        skip_errors = self.py_args["skip_errors"]
        rst_literal_blocks = self.py_args["rst_literal_blocks"]
        check_only = self.py_args["check_only"]

        # Build filepaths.
        exts = ("md", "py") if exts is None else tuple(ext.lstrip(".") for ext in exts)
        if exclude:
            exclude = tuple(re.compile(regex, re.I) if isinstance(regex, str) else regex for regex in exclude)
        filepaths = set()
        for path in paths:
            path = Path(path)  # noqa: PLW2901
            if path.is_file():
                filepaths.add(path.as_posix())
            else:
                for ext in exts:
                    filepaths |= {filepath.as_posix() for filepath in path.rglob(f"*.{ext}")}

        # Initiate black.
        black_mode = black.Mode(
            line_length=line_length or black.DEFAULT_LINE_LENGTH,
            string_normalization=string_normalization,
            is_pyi=is_pyi,
            is_ipynb=is_ipynb,
            skip_source_first_line=skip_source_first_line,
            magic_trailing_comma=magic_trailing_comma,
            python_cell_magics=python_cell_magics or set(),
            preview=preview,
        )

        # Run blacken-docs.
        retv = 0
        for filepath in sorted(filepaths):
            retv |= format_file(
                filepath,
                black_mode,
                skip_errors=skip_errors,
                rst_literal_blocks=rst_literal_blocks,
                check_only=check_only,
            )
        return retv



# Utilities for creating tools.

from __future__ import annotations

import shlex
import sys
from io import StringIO
from typing import Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class LazyStdout(StringIO):
    """Lazy stdout buffer.

    Can be used when tools' main entry-points
    expect a file-like object for stdout.
    """

    def write(self, value: str) -> int:
        """Write a string to the stdout buffer."""
        return sys.stdout.write(value)

    def __repr__(self) -> str:
        return "stdout"


class LazyStderr(StringIO):
    """Lazy stderr buffer.

    Can be used when tools' main entry-points
    expect a file-like object for stderr.
    """

    def write(self, value: str) -> int:
        """Write a string to the stderr buffer."""
        return sys.stderr.write(value)

    def __repr__(self) -> str:
        return "stderr"


class Tool:
    """Base class for tools."""

    cli_name: str = ""
    """The name of the executable on PATH."""

    def __init__(
        self,
        cli_args: list[str] | None = None,
        py_args: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the tool.

        Parameters:
            cli_args: Initial command-line arguments. Use `add_args()` to add more.
            py_args: Python arguments. Your `__call__` method will be able to access
                these arguments as `self.py_args`.
        """
        self.cli_args: list[str] = cli_args or []
        """Registered command-line arguments."""
        self.py_args: dict[str, Any] = py_args or {}
        """Registered Python arguments."""

    def add_args(self, *args: str) -> Self:
        """Append CLI arguments."""
        self.cli_args.extend(args)
        return self

    @property
    def cli_command(self) -> str:
        """The equivalent CLI command."""
        if not self.cli_name:
            raise ValueError("This tool does not provide a CLI.")
        return shlex.join([self.cli_name, *self.cli_args])



from __future__ import annotations

import importlib
import sys
from io import StringIO
from typing import TYPE_CHECKING, Literal, cast

from duty._internal.tools._base import Tool

if TYPE_CHECKING:
    from collections.abc import Sequence


class safety(Tool):  # noqa: N801
    """Call [Safety](https://github.com/pyupio/safety)."""

    cli_name = "safety"
    """The name of the executable on PATH."""

    @classmethod
    def check(
        cls,
        requirements: str | Sequence[str],
        *,
        ignore_vulns: dict[str, str] | None = None,
        formatter: Literal["json", "bare", "text"] = "text",
        full_report: bool = True,
    ) -> safety:
        """Run the safety check command.

        This function makes sure we load the original, unpatched version of safety.

        Parameters:
            requirements: Python "requirements" (list of pinned dependencies).
            ignore_vulns: Vulnerabilities to ignore.
            formatter: Report format.
            full_report: Whether to output a full report.

        Returns:
            Success/failure.
        """
        return cls(py_args=dict(locals()))

    @property
    def cli_command(self) -> str:
        """The equivalent CLI command."""
        raise ValueError("This command cannot be translated to a CLI command.")

    def __call__(self) -> bool:
        """Run the command.

        Returns:
            False when vulnerabilities are found.
        """
        requirements = self.py_args["requirements"]
        ignore_vulns = self.py_args["ignore_vulns"]
        formatter = self.py_args["formatter"]
        full_report = self.py_args["full_report"]

        # set default parameter values
        ignore_vulns = ignore_vulns or {}

        # undo possible patching
        # see https://github.com/pyupio/safety/issues/348
        for module in sys.modules:
            if module.startswith("safety.") or module == "safety":
                del sys.modules[module]

        importlib.invalidate_caches()

        # reload original, unpatched safety
        from safety.formatter import SafetyFormatter  # noqa: PLC0415
        from safety.safety import calculate_remediations, check  # noqa: PLC0415
        from safety.util import read_requirements  # noqa: PLC0415

        # check using safety as a library
        if isinstance(requirements, (list, tuple, set)):
            requirements = "\n".join(requirements)
        packages = list(read_requirements(StringIO(cast("str", requirements))))

        # TODO: Safety 3 support, merge once support for v2 is dropped.
        check_kwargs = {"packages": packages, "ignore_vulns": ignore_vulns}
        try:
            from safety.auth.cli_utils import build_client_session  # noqa: PLC0415

            client_session, _ = build_client_session()
            check_kwargs["session"] = client_session
        except ImportError:
            pass

        vulns, db_full = check(**check_kwargs)
        remediations = calculate_remediations(vulns, db_full)
        output_report = SafetyFormatter(formatter).render_vulnerabilities(
            announcements=[],
            vulnerabilities=vulns,
            remediations=remediations,
            full=full_report,
            packages=packages,
        )

        # print report, return status
        if vulns:
            print(output_report)  # noqa: T201
            return False
        return True



from __future__ import annotations

from typing import Literal

from duty._internal.tools._base import Tool


class pytest(Tool):  # noqa: N801
    """Call [pytest](https://github.com/pytest-dev/pytest)."""

    cli_name = "pytest"
    """The name of the executable on PATH."""

    def __init__(
        self,
        *paths: str,
        config_file: str | None = None,
        select: str | None = None,
        select_markers: str | None = None,
        markers: bool | None = None,
        exitfirst: bool | None = None,
        fixtures: bool | None = None,
        fixtures_per_test: bool | None = None,
        pdb: bool | None = None,
        pdbcls: str | None = None,
        trace: bool | None = None,
        capture: str | None = None,
        runxfail: bool | None = None,
        last_failed: bool | None = None,
        failed_first: bool | None = None,
        new_first: bool | None = None,
        cache_show: str | None = None,
        cache_clear: bool | None = None,
        last_failed_no_failures: Literal["all", "none"] | None = None,
        stepwise: bool | None = None,
        stepwise_skip: bool | None = None,
        durations: int | None = None,
        durations_min: int | None = None,
        verbose: bool | None = None,
        no_header: bool | None = None,
        no_summary: bool | None = None,
        quiet: bool | None = None,
        verbosity: int | None = None,
        show_extra_summary: str | None = None,
        disable_pytest_warnings: bool | None = None,
        showlocals: bool | None = None,
        no_showlocals: bool | None = None,
        traceback: Literal["auto", "long", "short", "line", "native", "no"] | None = None,
        show_capture: Literal["no", "stdout", "stderr", "log", "all"] | None = None,
        full_trace: bool | None = None,
        color: str | None = None,
        code_highlight: bool | None = None,
        pastebin: str | None = None,
        junit_xml: str | None = None,
        junit_prefix: str | None = None,
        pythonwarnings: str | None = None,
        maxfail: int | None = None,
        strict_config: bool | None = None,
        strict_markers: bool | None = None,
        continue_on_collection_errors: bool | None = None,
        rootdir: str | None = None,
        collect_only: bool | None = None,
        pyargs: bool | None = None,
        ignore: list[str] | None = None,
        ignore_glob: list[str] | None = None,
        deselect: str | None = None,
        confcutdir: str | None = None,
        noconftest: bool | None = None,
        keep_duplicates: bool | None = None,
        collect_in_virtualenv: bool | None = None,
        import_mode: Literal["prepend", "append", "importlib"] | None = None,
        doctest_modules: bool | None = None,
        doctest_report: Literal["none", "cdiff", "ndiff", "udiff", "only_first_failure"] | None = None,
        doctest_glob: str | None = None,
        doctest_ignore_import_errors: bool | None = None,
        doctest_continue_on_failure: bool | None = None,
        basetemp: str | None = None,
        plugins: list[str] | None = None,
        no_plugins: list[str] | None = None,
        trace_config: bool | None = None,
        debug: str | None = None,
        override_ini: str | None = None,
        assert_mode: str | None = None,
        setup_only: bool | None = None,
        setup_show: bool | None = None,
        setup_plan: bool | None = None,
        log_level: str | None = None,
        log_format: str | None = None,
        log_date_format: str | None = None,
        log_cli_level: tuple[str, str] | None = None,
        log_cli_format: str | None = None,
        log_cli_date_format: str | None = None,
        log_file: str | None = None,
        log_file_level: str | None = None,
        log_file_format: str | None = None,
        log_file_date_format: str | None = None,
        log_auto_indent: str | None = None,
    ) -> None:
        """Run `pytest`.

        Parameters:
            *paths: Files or directories to select tests from.
            select: Only run tests which match the given substring expression. An expression is a Python evaluatable expression where all names are substring-matched against test names and their parent classes. Example: -k 'test_method or test_other' matches all test functions and classes whose name contains 'test_method' or 'test_other', while -k 'not test_method' matches those that don't contain 'test_method' in their names. -k 'not test_method and not test_other' will eliminate the matches. Additionally keywords are matched to classes and functions containing extra names in their 'extra_keyword_matches' set, as well as functions which have names assigned directly to them. The matching is case-insensitive.
            select_markers: Only run tests matching given mark expression. For example: -m 'mark1 and not mark2'.
            markers: show markers (builtin, plugin and per-project ones).
            exitfirst: Exit instantly on first error or failed test
            fixtures: Show available fixtures, sorted by plugin appearance (fixtures with leading '_' are only shown with '-v')
            fixtures_per_test: Show fixtures per test
            pdb: Start the interactive Python debugger on errors or KeyboardInterrupt
            pdbcls: Specify a custom interactive Python debugger for use with --pdb.For example: --pdbcls IPython.terminal.debugger:TerminalPdb
            trace: Immediately break when running each test
            capture: Per-test capturing method: one of fd|sys|no|tee-sys
            runxfail: Report the results of xfail tests as if they were not marked
            last_failed: Rerun only the tests that failed at the last run (or all if none failed)
            failed_first: Run all tests, but run the last failures first. This may re-order tests and thus lead to repeated fixture setup/teardown.
            new_first: Run tests from new files first, then the rest of the tests sorted by file mtime
            cache_show: Show cache contents, don't perform collection or tests. Optional argument: glob (default: '*').
            cache_clear: Remove all cache contents at start of test run
            last_failed_no_failures: Which tests to run with no previously (known) failures
            stepwise: Exit on test failure and continue from last failing test next time
            stepwise_skip: Ignore the first failing test but stop on the next failing test. Implicitly enables --stepwise.
            durations: Show N slowest setup/test durations (N 0 for all)
            durations_min: Minimal duration in seconds for inclusion in slowest list. Default: 0.005.
            verbose: Increase verbosity
            no_header: Disable header
            no_summary: Disable summary
            quiet: Decrease verbosity
            verbosity: Set verbosity. Default: 0.
            show_extra_summary: Show extra test summary info as specified by chars: (f)ailed, (E)rror, (s)kipped, (x)failed, (X)passed, (p)assed, (P)assed with output, (a)ll except passed (p/P), or (A)ll. (w)arnings are enabled by default (see --disable-warnings), 'N' can be used to reset the list. (default: 'fE').
            disable_pytest_warnings: Disable warnings summary
            showlocals: Show locals in tracebacks (disabled by default)
            no_showlocals: Hide locals in tracebacks (negate --showlocals passed through addopts)
            traceback: Traceback print mode (auto/long/short/line/native/no)
            show_capture: Controls how captured stdout/stderr/log is shown on failed tests. Default: all.
            full_trace: Don't cut any tracebacks (default is to cut)
            color: Color terminal output (yes/no/auto)
            code_highlight: {yes,no} Whether code should be highlighted (only if --color is also enabled). Default: yes.
            pastebin: Send failed|all info to bpaste.net pastebin service
            junit_xml: Create junit-xml style report file at given path
            junit_prefix: Prepend prefix to classnames in junit-xml output
            pythonwarnings: Set which warnings to report, see -W option of Python itself
            maxfail: Exit after first num failures or errors
            strict_config: Any warnings encountered while parsing the `pytest` section of the configuration file raise errors
            strict_markers: Markers not registered in the `markers` section of the configuration file raise errors
            config_file: Load configuration from `file` instead of trying to locate one of the implicit configuration files
            continue_on_collection_errors: Force test execution even if collection errors occur
            rootdir: Define root directory for tests. Can be relative path: 'root_dir', './root_dir', 'root_dir/another_dir/'; absolute path: '/home/user/root_dir'; path with variables: '$HOME/root_dir'.
            collect_only: Only collect tests, don't execute them
            pyargs: Try to interpret all arguments as Python packages
            ignore: Ignore path during collection (multi-allowed)
            ignore_glob: Ignore path pattern during collection (multi-allowed)
            deselect: Deselect item (via node id prefix) during collection (multi-allowed)
            confcutdir: Only load conftest.py's relative to specified dir
            noconftest: Don't load any conftest.py files
            keep_duplicates: Keep duplicate tests
            collect_in_virtualenv: Don't ignore tests in a local virtualenv directory
            import_mode: Prepend/append to sys.path when importing test modules and conftest files. Default: prepend.
            doctest_modules: Run doctests in all .py modules
            doctest_report: Choose another output format for diffs on doctest failure
            doctest_glob: Doctests file matching pattern, default: test*.txt
            doctest_ignore_import_errors: Ignore doctest ImportErrors
            doctest_continue_on_failure: For a given doctest, continue to run after the first failure
            basetemp: Base temporary directory for this test run. (Warning: this directory is removed if it exists.)
            plugins: Early-load given plugin module name or entry point (multi-allowed). To avoid loading of plugins, use the `no:` prefix, e.g. `no:doctest`.
            no_plugins: Early-load given plugin module name or entry point (multi-allowed). To avoid loading of plugins, use the `no:` prefix, e.g. `no:doctest`.
            trace_config: Trace considerations of conftest.py files
            debug: Store internal tracing debug information in this log file. This file is opened with 'w' and truncated as a result, care advised. Default: pytestdebug.log.
            override_ini: Override ini option with "option value" style, e.g. `-o xfail_strict True -o cache_dir cache`.
            assert_mode: Control assertion debugging tools. 'plain' performs no assertion debugging. 'rewrite' (the default) rewrites assert statements in test modules on import to provide assert expression information.
            setup_only: Only setup fixtures, do not execute tests
            setup_show: Show setup of fixtures while executing tests
            setup_plan: Show what fixtures and tests would be executed but don't execute anything
            log_level: Level of messages to catch/display. Not set by default, so it depends on the root/parent log handler's effective level, where it is "WARNING" by default.
            log_format: Log format used by the logging module.
            log_date_format: Log date format used by the logging module.
            log_cli_level: logging level.
            log_cli_format: Log format used by the logging module.
            log_cli_date_format: Log date format used by the logging module.
            log_file: Path to a file when logging will be written to.
            log_file_level: Log file logging level.
            log_file_format: Log format used by the logging module.
            log_file_date_format: Log date format used by the logging module.
            log_auto_indent: Auto-indent multiline messages passed to the logging module. Accepts true|on, false|off or an integer.
        """
        cli_args = list(paths)

        if select:
            cli_args.append("-k")
            cli_args.append(select)

        if select_markers:
            cli_args.append("-m")
            cli_args.append(select_markers)

        if markers:
            cli_args.append("--markers")

        if exitfirst:
            cli_args.append("--exitfirst")

        if fixtures:
            cli_args.append("--fixtures")

        if fixtures_per_test:
            cli_args.append("--fixtures-per-test")

        if pdb:
            cli_args.append("--pdb")

        if pdbcls:
            cli_args.append("--pdbcls")
            cli_args.append(pdbcls)

        if trace:
            cli_args.append("--trace")

        if capture:
            cli_args.append("--capture")

        if runxfail:
            cli_args.append("--runxfail")

        if last_failed:
            cli_args.append("--last-failed")

        if failed_first:
            cli_args.append("--failed-first")

        if new_first:
            cli_args.append("--new-first")

        if cache_show:
            cli_args.append("--cache-show")
            cli_args.append(cache_show)

        if cache_clear:
            cli_args.append("--cache-clear")

        if last_failed_no_failures:
            cli_args.append("--last-failed-no-failures")
            cli_args.append(last_failed_no_failures)

        if stepwise:
            cli_args.append("--stepwise")

        if stepwise_skip:
            cli_args.append("--stepwise-skip")

        if durations:
            cli_args.append("--durations")
            cli_args.append(str(durations))

        if durations_min:
            cli_args.append("--durations-min")
            cli_args.append(str(durations_min))

        if verbose:
            cli_args.append("--verbose")

        if no_header:
            cli_args.append("--no-header")

        if no_summary:
            cli_args.append("--no-summary")

        if quiet:
            cli_args.append("--quiet")

        if verbosity:
            cli_args.append("--verbosity")
            cli_args.append(str(verbosity))

        if show_extra_summary:
            cli_args.append("-r")
            cli_args.append(show_extra_summary)

        if disable_pytest_warnings:
            cli_args.append("--disable-pytest-warnings")

        if showlocals:
            cli_args.append("--showlocals")

        if no_showlocals:
            cli_args.append("--no-showlocals")

        if traceback:
            cli_args.append("--tb")
            cli_args.append(traceback)

        if show_capture:
            cli_args.append("--show-capture")
            cli_args.append(show_capture)

        if full_trace:
            cli_args.append("--full-trace")

        if color:
            cli_args.append("--color")
            cli_args.append(color)

        if code_highlight:
            cli_args.append("--code-highlight")

        if pastebin:
            cli_args.append("--pastebin")
            cli_args.append(pastebin)

        if junit_xml:
            cli_args.append("--junit-xml")
            cli_args.append(junit_xml)

        if junit_prefix:
            cli_args.append("--junit-prefix")
            cli_args.append(junit_prefix)

        if pythonwarnings:
            cli_args.append("--pythonwarnings")
            cli_args.append(pythonwarnings)

        if maxfail:
            cli_args.append("--maxfail")
            cli_args.append(str(maxfail))

        if strict_config:
            cli_args.append("--strict-config")

        if strict_markers:
            cli_args.append("--strict-markers")

        if config_file:
            cli_args.append("-c")
            cli_args.append(config_file)

        if continue_on_collection_errors:
            cli_args.append("--continue-on-collection-errors")

        if rootdir:
            cli_args.append("--rootdir")
            cli_args.append(rootdir)

        if collect_only:
            cli_args.append("--collect-only")

        if pyargs:
            cli_args.append("--pyargs")

        if ignore:
            for ign in ignore:
                cli_args.append("--ignore")
                cli_args.append(ign)

        if ignore_glob:
            for ign_glob in ignore_glob:
                cli_args.append("--ignore-glob")
                cli_args.append(ign_glob)

        if deselect:
            cli_args.append("--deselect")
            cli_args.append(deselect)

        if confcutdir:
            cli_args.append("--confcutdir")
            cli_args.append(confcutdir)

        if noconftest:
            cli_args.append("--noconftest")

        if keep_duplicates:
            cli_args.append("--keep-duplicates")

        if collect_in_virtualenv:
            cli_args.append("--collect-in-virtualenv")

        if import_mode:
            cli_args.append("--import-mode")
            cli_args.append(import_mode)

        if doctest_modules:
            cli_args.append("--doctest-modules")

        if doctest_report:
            cli_args.append("--doctest-report")
            cli_args.append(doctest_report)

        if doctest_glob:
            cli_args.append("--doctest-glob")
            cli_args.append(doctest_glob)

        if doctest_ignore_import_errors:
            cli_args.append("--doctest-ignore-import-errors")

        if doctest_continue_on_failure:
            cli_args.append("--doctest-continue-on-failure")

        if basetemp:
            cli_args.append("--basetemp")
            cli_args.append(basetemp)

        if plugins:
            for plugin in plugins:
                cli_args.append("-p")
                cli_args.append(plugin)

        if no_plugins:
            for no_plugin in no_plugins:
                cli_args.append("-p")
                cli_args.append(f"no:{no_plugin}")

        if trace_config:
            cli_args.append("--trace-config")

        if debug:
            cli_args.append("--debug")
            cli_args.append(debug)

        if override_ini:
            cli_args.append("--override-ini")
            cli_args.append(override_ini)

        if assert_mode:
            cli_args.append("--assert")
            cli_args.append(assert_mode)

        if setup_only:
            cli_args.append("--setup-only")

        if setup_show:
            cli_args.append("--setup-show")

        if setup_plan:
            cli_args.append("--setup-plan")

        if log_level:
            cli_args.append("--log-level")
            cli_args.append(log_level)

        if log_format:
            cli_args.append("--log-format")
            cli_args.append(log_format)

        if log_date_format:
            cli_args.append("--log-date-format")
            cli_args.append(log_date_format)

        if log_cli_level:
            cli_args.append("--log-cli-level")
            cli_args.extend(log_cli_level)

        if log_cli_format:
            cli_args.append("--log-cli-format")
            cli_args.append(log_cli_format)

        if log_cli_date_format:
            cli_args.append("--log-cli-date-format")
            cli_args.append(log_cli_date_format)

        if log_file:
            cli_args.append("--log-file")
            cli_args.append(log_file)

        if log_file_level:
            cli_args.append("--log-file-level")
            cli_args.append(log_file_level)

        if log_file_format:
            cli_args.append("--log-file-format")
            cli_args.append(log_file_format)

        if log_file_date_format:
            cli_args.append("--log-file-date-format")
            cli_args.append(log_file_date_format)

        if log_auto_indent:
            cli_args.append("--log-auto-indent")
            cli_args.append(log_auto_indent)

        super().__init__(cli_args)

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        from pytest import main as run_pytest  # noqa: PT013,PLC0415

        return run_pytest(self.cli_args)



"""Tests for the `decorator` module."""

from __future__ import annotations

import inspect

import pytest

from duty._internal.context import Context
from duty._internal.decorator import duty as decorate
from duty._internal.exceptions import DutyFailure


def test_accept_one_posarg_when_decorating() -> None:
    """Accept only one positional argument when decorating."""
    with pytest.raises(ValueError, match="accepts only one positional argument"):
        decorate(0, 1)  # type: ignore[call-overload]


def test_skipping() -> None:
    """Wrap function that must be skipped."""
    duty = decorate(lambda ctx: ctx.run("false"), skip_if=True)  # type: ignore[call-overload]
    # no DutyFailure raised
    assert duty.run() is None
    with pytest.raises(DutyFailure):
        assert inspect.unwrap(duty)(Context({}))



"""Generate the credits page."""

import functools
import re
import urllib
from itertools import chain
from pathlib import Path

import mkdocs_gen_files
import toml
from jinja2 import StrictUndefined
from jinja2.sandbox import SandboxedEnvironment


def get_credits_data() -> dict:
    """Return data used to generate the credits file.

    Returns:
        Data required to render the credits template.
    """
    project_dir = Path(__file__).parent.parent
    metadata = toml.load(project_dir / "pyproject.toml")["project"]
    metadata_pdm = toml.load(project_dir / "pyproject.toml")["tool"]["pdm"]
    lock_data = toml.load(project_dir / "pdm.lock")
    project_name = metadata["name"]

    all_dependencies = chain(
        metadata.get("dependencies", []),
        chain(*metadata.get("optional-dependencies", {}).values()),
        chain(*metadata_pdm.get("dev-dependencies", {}).values()),
    )
    direct_dependencies = {re.sub(r"[^\w-].*$", "", dep) for dep in all_dependencies}
    direct_dependencies = {dep.lower() for dep in direct_dependencies}
    indirect_dependencies = {pkg["name"].lower() for pkg in lock_data["package"]}
    indirect_dependencies -= direct_dependencies

    return {
        "project_name": project_name,
        "direct_dependencies": sorted(direct_dependencies),
        "indirect_dependencies": sorted(indirect_dependencies),
        "more_credits": "http://pawamoy.github.io/credits/",
    }


@functools.lru_cache(maxsize=None)
def get_credits():
    """Return credits as Markdown.

    Returns:
        The credits page Markdown.
    """
    jinja_env = SandboxedEnvironment(undefined=StrictUndefined)
    commit = "398879aba2a365049870709116a689618afeb5b7"
    template_url = f"https://raw.githubusercontent.com/pawamoy/jinja-templates/{commit}/credits.md"
    template_data = get_credits_data()
    template_text = urllib.request.urlopen(template_url).read().decode("utf8")  # noqa: S310
    return jinja_env.from_string(template_text).render(**template_data)


with mkdocs_gen_files.open("credits.md", "w") as fd:
    fd.write(get_credits())
mkdocs_gen_files.set_edit_path("credits.md", "gen_credits.py")



"""Deprecated. Use [`duty.tools.ssort`][] instead."""

# YORE: Bump 2: Remove file.

import warnings
from typing import Any

from duty._internal.callables import ssort as _ssort


def __getattr__(name: str) -> Any:
    warnings.warn(
        "Callables are deprecated in favor of tools, use `duty.tools.ssort` instead of `duty.callables.ssort`.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(_ssort, name)



"""duty package.

A simple task runner.
"""

from __future__ import annotations

from failprint import lazy

from duty._internal.cli import (
    empty,
    get_duty_parser,
    get_parser,
    main,
    parse_args,
    parse_commands,
    parse_options,
    print_help,
    specified_options,
    split_args,
)
from duty._internal.collection import Collection, Duty, DutyListType, default_duties_file
from duty._internal.context import CmdType, Context
from duty._internal.decorator import create_duty, duty
from duty._internal.exceptions import DutyFailure
from duty._internal.tools._base import LazyStderr, LazyStdout, Tool
from duty._internal.validation import ParamsCaster, cast_arg, to_bool, validate

__all__: list[str] = [
    "CmdType",
    "Collection",
    "Context",
    "Duty",
    "DutyFailure",
    "DutyListType",
    "LazyStderr",
    "LazyStdout",
    "ParamsCaster",
    "Tool",
    "cast_arg",
    "create_duty",
    "default_duties_file",
    "duty",
    "empty",
    "get_duty_parser",
    "get_parser",
    "lazy",
    "main",
    "parse_args",
    "parse_commands",
    "parse_options",
    "print_help",
    "specified_options",
    "split_args",
    "to_bool",
    "validate",
]



from __future__ import annotations

import os
import subprocess
from functools import cache

from duty._internal.tools._base import Tool


@cache
def _find_ty() -> str:
    from ty.__main__ import find_ty_bin  # noqa: PLC0415

    try:
        return find_ty_bin()
    except FileNotFoundError:
        paths = os.environ["PATH"]
        for path in paths.split(os.pathsep):
            ty = os.path.join(path, "ty")
            if os.path.exists(ty):
                return ty
    return "ty"


class ty(Tool):  # noqa: N801
    """Call [ty](https://github.com/astral-sh/ty)."""

    cli_name = "ty"
    """The name of the executable on PATH."""

    @classmethod
    def check(
        cls,
        *paths: str,
        add_ignore: bool | None = None,
        project: str | None = None,
        python: str | None = None,
        typeshed: str | None = None,
        extra_search_path: list[str] | None = None,
        python_version: str | None = None,
        python_platform: str | None = None,
        verbose: bool = False,
        quiet: bool = False,
        config: list[str] | None = None,
        config_file: str | None = None,
        output_format: str | None = None,
        error_on_warning: bool | None = None,
        exit_zero: bool | None = None,
        watch: bool | None = None,
        error: list[str] | None = None,
        warn: list[str] | None = None,
        ignore: list[str] | None = None,
        respect_ignore_files: bool | None = None,
        force_exclude: bool | None = None,
        exclude: list[str] | None = None,
        no_progress: bool | None = None,
        color: str | bool | None = None,
    ) -> ty:
        """Run ty on the given files or directories.

        Parameters:
            paths: Files or directories to type check.
            add_ignore: Adds `ty: ignore` comments to suppress all rule diagnostics.
            project: Run the command within the given project directory.
            python: Path to your project's Python environment or interpreter.
            typeshed: Custom directory to use for stdlib typeshed stubs.
            extra_search_path: Additional path to use as a module-resolution source (can be passed multiple times).
            python_version: Python version to assume when resolving types (possible values: 3.7, 3.8, 3.9, 3.10, 3.11, 3.12, 3.13, 3.14).
            python_platform: Target platform to assume when resolving types.
            verbose: Use verbose output (or `-vv` and `-vvv` for more verbose output).
            quiet: Use quiet output (or `-qq` for silent output).
            config: A TOML `<KEY> = <VALUE>` pair overriding a specific configuration option.
            config_file: The path to a `ty.toml` file to use for configuration.
            output_format: The format to use for printing diagnostic messages (possible values: full, concise, gitlab, github).
            error_on_warning: Use exit code 1 if there are any warning-level diagnostics.
            exit_zero: Always use exit code 0, even when there are error-level diagnostics.
            watch: Watch files for changes and recheck files related to the changed files.
            error: Treat the given rule as having severity 'error'. Can be specified multiple times.
            warn: Treat the given rule as having severity 'warn'. Can be specified multiple times.
            ignore: Disables the rule. Can be specified multiple times.
            respect_ignore_files: Respect file exclusions via `.gitignore` and other standard ignore files.
            force_exclude: Enforce exclusions, even for paths passed to ty directly on the command-line.
            exclude: Glob patterns for files to exclude from type checking.
            no_progress: Hide all progress outputs.
            color: Control when colored output is used (possible values: auto, always/true, never/false).
        """
        cli_args = ["check", *paths]

        if add_ignore:
            cli_args.append("--add-ignore")

        if project:
            cli_args.append("--project")
            cli_args.append(project)

        if python:
            cli_args.append("--python")
            cli_args.append(python)

        if typeshed:
            cli_args.append("--typeshed")
            cli_args.append(typeshed)

        if extra_search_path:
            for path in extra_search_path:
                cli_args.append("--extra-search-path")
                cli_args.append(path)

        if python_version:
            cli_args.append("--python-version")
            cli_args.append(python_version)

        if python_platform:
            cli_args.append("--python-platform")
            cli_args.append(python_platform)

        if verbose:
            cli_args.append("--verbose")

        if quiet:
            cli_args.append("--quiet")

        if config:
            for config_option in config:
                cli_args.append("--config")
                cli_args.append(config_option)

        if config_file:
            cli_args.append("--config-file")
            cli_args.append(config_file)

        if output_format:
            cli_args.append("--output-format")
            cli_args.append(output_format)

        if error_on_warning:
            cli_args.append("--error-on-warning")

        if exit_zero:
            cli_args.append("--exit-zero")

        if watch:
            cli_args.append("--watch")

        if error:
            for rule in error:
                cli_args.append("--error")
                cli_args.append(rule)

        if warn:
            for rule in warn:
                cli_args.append("--warn")
                cli_args.append(rule)

        if ignore:
            for rule in ignore:
                cli_args.append("--ignore")
                cli_args.append(rule)

        if respect_ignore_files:
            cli_args.append("--respect-ignore-files")

        if force_exclude:
            cli_args.append("--force-exclude")

        if exclude:
            for pattern in exclude:
                cli_args.append("--exclude")
                cli_args.append(pattern)

        if no_progress:
            cli_args.append("--no-progress")

        if color is True:
            cli_args.append("--color")
            cli_args.append("always")
        elif color is False:
            cli_args.append("--color")
            cli_args.append("never")

        return cls(cli_args)

    def __call__(self) -> int:
        """Run the command.

        Returns:
            The exit code of the command.
        """
        process = subprocess.run(  # noqa: S603
            [_find_ty(), *self.cli_args],
            capture_output=True,
            text=True,
            check=False,
        )
        print(process.stdout)  # noqa: T201
        return process.returncode



# This module contains logic used to validate parameters passed to duties.
#
# We validate the parameters before running the duties,
# effectively checking all CLI arguments and failing early
# if they are incorrect.

from __future__ import annotations

import sys
import textwrap
from contextlib import suppress
from functools import cached_property, partial
from inspect import Parameter, Signature, signature
from types import UnionType
from typing import (  # type: ignore[attr-defined]
    TYPE_CHECKING,
    Any,
    Callable,
    ForwardRef,
    Union,
    _eval_type,
    get_args,
    get_origin,
)

if sys.version_info >= (3, 13):
    _eval_type = partial(_eval_type, type_params=None)
_union_types = (Union, UnionType)

if TYPE_CHECKING:
    from collections.abc import Sequence


def to_bool(value: str) -> bool:
    """Convert a string to a boolean.

    Parameters:
        value: The string to convert.

    Returns:
        True or False.
    """
    return value.lower() not in {"", "0", "no", "n", "false", "off"}


def cast_arg(arg: Any, annotation: Any) -> Any:
    """Cast an argument using a type annotation.

    Parameters:
        arg: The argument value.
        annotation: A type annotation.

    Returns:
        The cast value.
    """
    if annotation is Parameter.empty:
        return arg
    if annotation is bool:
        annotation = to_bool
    if get_origin(annotation) in _union_types:
        for sub_annotation in get_args(annotation):
            if sub_annotation is type(None):
                continue
            with suppress(Exception):
                return cast_arg(arg, sub_annotation)
    try:
        return annotation(arg)
    except Exception:  # noqa: BLE001
        return arg


class ParamsCaster:
    """A helper class to cast parameters based on a function's signature annotations."""

    def __init__(self, signature: Signature) -> None:
        """Initialize the object.

        Parameters:
            signature: The signature to use to cast arguments.
        """
        self.params_dict = signature.parameters
        """A dictionary of parameters, indexed by their name."""
        self.params_list = list(self.params_dict.values())
        """A list of parameters, in the order they appear in the signature."""

    @cached_property
    def var_positional_position(self) -> int:
        """Give the position of the variable positional parameter in the signature.

        Returns:
            The position of the variable positional parameter.
        """
        for pos, param in enumerate(self.params_list):
            if param.kind is Parameter.VAR_POSITIONAL:
                return pos
        return -1

    @cached_property
    def has_var_positional(self) -> bool:
        """Tell if there is a variable positional parameter.

        Returns:
            True or False.
        """
        return self.var_positional_position >= 0

    @cached_property
    def var_positional_annotation(self) -> Any:
        """Give the variable positional parameter (`*args`) annotation if any.

        Returns:
            The variable positional parameter annotation.
        """
        return self.params_list[self.var_positional_position].annotation

    @cached_property
    def var_keyword_annotation(self) -> Any:
        """Give the variable keyword parameter (`**kwargs`) annotation if any.

        Returns:
            The variable keyword parameter annotation.
        """
        for param in self.params_list:
            if param.kind is Parameter.VAR_KEYWORD:
                return param.annotation
        return Parameter.empty

    def annotation_at_pos(self, pos: int) -> Any:
        """Give the annotation for the parameter at the given position.

        Parameters:
            pos: The position of the parameter.

        Returns:
            The positional parameter annotation.
        """
        return self.params_list[pos].annotation

    def eaten_by_var_positional(self, pos: int) -> bool:
        """Tell if the parameter at this position is eaten by a variable positional parameter.

        Parameters:
            pos: The position of the parameter.

        Returns:
            Whether the parameter is eaten.
        """
        return self.has_var_positional and pos >= self.var_positional_position

    def cast_posarg(self, pos: int, arg: Any) -> Any:
        """Cast a positional argument.

        Parameters:
            pos: The position of the argument in the signature.
            arg: The argument value.

        Returns:
            The cast value.
        """
        if self.eaten_by_var_positional(pos):
            return cast_arg(arg, self.var_positional_annotation)
        return cast_arg(arg, self.annotation_at_pos(pos))

    def cast_kwarg(self, name: str, value: Any) -> Any:
        """Cast a keyword argument.

        Parameters:
            name: The name of the argument in the signature.
            value: The argument value.

        Returns:
            The cast value.
        """
        if name in self.params_dict:
            return cast_arg(value, self.params_dict[name].annotation)
        return cast_arg(value, self.var_keyword_annotation)

    def cast(self, *args: Any, **kwargs: Any) -> tuple[Sequence, dict[str, Any]]:
        """Cast all positional and keyword arguments.

        Parameters:
            *args: The positional arguments.
            **kwargs: The keyword arguments.

        Returns:
            The cast arguments.
        """
        positional = tuple(self.cast_posarg(pos, arg) for pos, arg in enumerate(args))
        keyword = {name: self.cast_kwarg(name, value) for name, value in kwargs.items()}
        return positional, keyword


def _get_params_caster(func: Callable, *args: Any, **kwargs: Any) -> ParamsCaster:
    duties_module = sys.modules[func.__module__]
    exec_globals = dict(duties_module.__dict__)
    eval_str = False
    for name in list(exec_globals.keys()):
        if exec_globals[name] is annotations:
            eval_str = True
            del exec_globals[name]
            break
    exec_globals["__context_above"] = {}

    # Don't keep first parameter: context.
    params = list(signature(func).parameters.values())[1:]
    params_no_types = [Parameter(param.name, param.kind, default=param.default) for param in params]
    code_sig = Signature(parameters=params_no_types)
    if eval_str:
        params_types = [
            Parameter(
                param.name,
                param.kind,
                default=param.default,
                annotation=(
                    _eval_type(
                        ForwardRef(param.annotation) if isinstance(param.annotation, str) else param.annotation,
                        exec_globals,
                        {},
                    )
                    if param.annotation is not Parameter.empty
                    else type(param.default)
                ),
            )
            for param in params
        ]
    else:
        params_types = params
    cast_sig = Signature(parameters=params_types)

    code = f"""
        import inspect
        def {func.__name__}{code_sig}: ...
        __context_above['func'] = {func.__name__}
    """

    exec(textwrap.dedent(code), exec_globals)  # noqa: S102
    func = exec_globals["__context_above"]["func"]

    # Trigger TypeError early.
    func(*args, **kwargs)

    return ParamsCaster(cast_sig)


def validate(
    func: Callable,
    *args: Any,
    **kwargs: Any,
) -> tuple[Sequence, dict[str, Any]]:
    """Validate positional and keyword arguments against a function.

    First we clone the function, removing the first parameter (the context)
    and the body, to fail early with a `TypeError` if the arguments
    are incorrect: not enough, too much, in the wrong order, etc.

    Then we cast all the arguments using the function's signature
    and we return them.

    Parameters:
        func: The function to copy.
        *args: The positional arguments.
        **kwargs: The keyword arguments.

    Returns:
        The casted arguments.
    """
    return _get_params_caster(func, *args, **kwargs).cast(*args, **kwargs)



"""Deprecated. Use [`duty.tools.blacken_docs`][] instead."""

# YORE: Bump 2: Remove file.

import warnings
from typing import Any

from duty._internal.callables import blacken_docs as _blacken_docs


def __getattr__(name: str) -> Any:
    warnings.warn(
        "Callables are deprecated in favor of tools, use `duty.tools.blacken_docs` instead of `duty.callables.blacken_docs`.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(_blacken_docs, name)
